//! Сборник типов для внутренней обработки данных с датчиков.
//!
//! Предполагается, что struct-ы будут совместимы с JSON API для передачи данных датчиков.

// Не забывайте про:
// #[serde(rename = "...")]
// #[serde(alias = "...")]

use crate::ingest_protocol::error::Error;
use crate::ingest_protocol::parser::parse_mac_address;
use crate::uformat;
use crate::utils::{EpochUTC, SupportedUnit};
use crate::web_server::app_error::{AppError, ServerRedisSnafu};

use fred::clients::Client as RedisClient;
use fred::interfaces::{HashesInterface, KeysInterface};
use hifitime::Epoch;
use rust_decimal::Decimal;
use serde::{Deserialize, Serialize};
use serde_with::serde_as;
use snafu::ResultExt;

use std::collections::{HashMap, HashSet};
use std::hash::{Hash, Hasher};

/// Данные с одного датчика.
///
/// Основной идентификатор - поле `mac`. В случае с [crate::web_server::old_device_sensor_api],
/// `mac` может быть не полем, а названием ключа в параметрах. Из-за двусмысленности документации
/// NarodMon, `mac` может означать EUI-48 совместимый MAC адрес, или же просто
/// уникальный идентификатор.
///
/// Парсинг этих данных отличается в разных транспортных протоколах.
/// Для HTTP /post или /get: см. [crate::web_server::old_device_sensor_api]
/// Для TCP/UDP: TODO
/// Для MQTT: TODO
#[derive(Debug, Clone, Default, PartialEq, Eq, Serialize, Deserialize)]
pub struct SensorValue {
    pub mac: String,
    pub value: Decimal,
    pub time: Option<EpochUTC>,
    pub unit: Option<SupportedUnit>,
    pub name: Option<String>,
}

impl Hash for SensorValue {
    fn hash<H: Hasher>(&self, state: &mut H) {
        self.mac.hash(state)
    }
}

/// Функция-помощник для [MacAsArray], предназначенный для использования с [serde_with].
///
/// Преобразует MAC-адрес.
fn mac_as_array(value: &str) -> Result<[u8; 6], Error<&str>> {
    Ok(parse_mac_address(value)?.1)
}

serde_with::serde_conv!(
    MacAsArray,
    [u8; 6],
    |rgb: &[u8; 6]| rgb.to_owned(),
    mac_as_array
);

#[serde_as]
#[derive(Debug, Clone, Default, PartialEq, Eq, Serialize, Deserialize)]
pub struct NMDeviceDataPacket {
    #[serde(alias = "ID")]
    #[serde_as(as = "MacAsArray")]
    pub mac: [u8; 6],

    pub name: Option<String>,

    pub values: HashSet<SensorValue>,

    pub owner: Option<String>,

    pub lat: Option<Decimal>,
    pub lon: Option<Decimal>,
    pub alt: Option<Decimal>,

    /// Команды для управления устройством.
    ///
    /// Текст после декодировки.
    ///
    /// ## Отчёт совместимости
    /// Функционал изначально narodmon-овский.
    /// Про то, как кодировать при наличии символов `#;=` или поддержку UTF-8 не сказано.
    /// Рассчитываю на то, что все значения нужно URL кодировать.
    pub commands: Option<HashMap<String, String>>,

    // HTTP GET/POST url-encode specific parameters
    /// TODO: Желательное поведение в будущем:
    ///   - Если в values есть хотябы один times и этот time не None => Игнорируем этот time
    ///   - Если time нет у values => используем этот time (если он None, то соотв. считаем что информации о времени нет)
    /// TODO: В базе всё должно храниться как секунды по TAI спустя J1900 для возможности сортировки по времени позже.
    pub time: Option<EpochUTC>,
}

impl NMDeviceDataPacket {
    pub async fn save_to_db(&self, redis: &RedisClient) -> Result<(), AppError> {
        let device_mac_enc = hex::encode(self.mac);

        let now = Epoch::now().unwrap().into();
        let mut device_time = self.time.unwrap_or(now);

        // TODO: Добавить гистерезис
        // Отчёт совместимости: отсутствует
        if device_time > now {
            device_time = now;
        }

        let device_tai_timestamp = device_time.0.to_duration_since_j1900().to_seconds();

        let key = uformat!("devices_{}", device_mac_enc);

        let device_exists: Option<bool> = redis
            .hget(key.as_str(), "exists")
            .await
            .context(ServerRedisSnafu)?;

        if !device_exists.is_some_and(|v| v) {
            return Err(AppError::DeviceNotFound {
                mac: hex::encode(self.mac),
            });
        }

        // devices_{device_id}_{tai_timestamp}_{sensor_id}
        for sensor in &self.values {
            let key = uformat!(
                "devices_{}_{}_{}",
                device_mac_enc,
                device_tai_timestamp.to_string(),
                sensor.mac
            );

            redis
                .set(key.as_str(), sensor.value.to_string(), None, None, false)
                .await
                .context(ServerRedisSnafu)?;
        }

        if let Some(commands) = &self.commands {
            for (cmd_key, cmd_value) in commands {
                let key = uformat!("devices_{}_cmds_{}", device_mac_enc, cmd_key);

                redis
                    .set(key.as_str(), cmd_value, None, None, false)
                    .await
                    .context(ServerRedisSnafu)?;
            }
        }

        Ok(())
    }
}

#[derive(Debug, Clone, Default, PartialEq, Deserialize)]
pub struct NMJsonPacket {
    pub devices: Vec<NMDeviceDataPacket>,
}

impl NMJsonPacket {
    pub async fn save_to_db(&self, redis: &RedisClient) -> Result<(), AppError> {
        for device in &self.devices {
            device.save_to_db(redis).await?;
        }

        Ok(())
    }
}
