use hifitime::Epoch;
use nom::bytes::complete::tag;
use nom::bytes::complete::take_until1;
use nom::bytes::complete::{take, take_while, take_while1};
use nom::character::complete::hex_digit1;
use std::str::FromStr;

use crate::ingest_protocol::error::Error;
use crate::ingest_protocol::{NMDeviceDataPacket, SensorValue};
use nom::combinator::{map_parser, opt};
use nom::error::context;
use nom::multi::{count, separated_list0};
use nom::sequence::{delimited, preceded};
use rust_decimal::Decimal;

type MyIError<I, O> = Result<(I, O), Error<I>>;

pub fn parse_mac_address(input: &str) -> MyIError<&str, [u8; 6]> {
    let mut mac = [0u8; 6];

    let mut counter = 0;

    let (leftovers, i) =
        context("17 символов для MAC адреса", take(17usize))(input)?;

    let (_, out) = count(
        |inp| {
            let (mut i, o) = context("Октет", map_parser(take(2usize), hex_digit1))(inp)?;
            if counter != 5 {
                (i, _) = tag("-")(i)?;
            }
            counter += 1;
            Ok((i, o))
        },
        6,
    )(i)?;

    hex::decode_to_slice(out.join(""), &mut mac).unwrap();

    Ok((leftovers, mac))
}

fn handle_special_sensor_macs<'a>(
    mac: &'a str,
    sensor_value: &str,
    packet: &mut NMDeviceDataPacket,
) -> Result<bool, Error<&'a str>> {
    match mac.to_uppercase().as_str() {
        "LAT" => packet.lat = Some(Decimal::from_str(sensor_value)?),
        "LON" => packet.lon = Some(Decimal::from_str(sensor_value)?),
        "ALT" => packet.alt = Some(Decimal::from_str(sensor_value)?),
        _ => {
            return Ok(false);
        }
    }
    Ok(true)
}

pub fn parse_packet_body<'a>(
    line: &'a str,
    packet: &mut NMDeviceDataPacket,
) -> MyIError<&'a str, ()> {
    let (line, _) = tag("#")(line)?;
    let (line, sensor_mac) = take_while1(|c| c != '\n' && c != '#')(line)?;

    let (line, _) = tag("#")(line)?;

    match sensor_mac {
        "OWNER" => {
            let (_line, owner_value) = take_while1(|c| c != '\n')(line)?;

            packet.owner = Some(owner_value.into())

            //hs.insert(NarodMonValues::Owner(owner_value.into()));
            //let (line, _) = tag("\n")(line)?;
        }
        _ => {
            let (line, sensor_value) = take_while1(|c| c != '\n' && c != '#')(line)?;
            let (line, sensor_time) = opt(preceded(
                tag("#"),
                take_while1(|c| c != '\n' && c != '#' && "1234567890.".contains(c)),
            ))(line)?;
            let (_line, sensor_name) = opt(preceded(tag("#"), take_while1(|c| c != '\n')))(line)?;

            let sensor_time = match sensor_time {
                Some(v) => Some(Epoch::from_unix_seconds(
                    v.parse().map_err(Error::TimestampParseError)?,
                )),
                None => None,
            };

            if !handle_special_sensor_macs(sensor_mac, sensor_value, packet)? {
                packet.values.insert(SensorValue {
                    mac: sensor_mac.into(),
                    value: Decimal::from_str(sensor_value)?,
                    time: sensor_time.map(|v| v.into()), // TODO
                    unit: None,
                    name: sensor_name.map(|v| v.to_string()),
                });
            }
        }
    }

    Ok((line, ()))
}

pub fn parse_packet(input: &str) -> MyIError<&str, NMDeviceDataPacket> {
    let (input, _) = tag("#")(input)?;

    let (input, device_mac) = parse_mac_address(input)?;

    let (input, opt_name) = opt(delimited(tag("#"), take_while(|c| c != '\n'), tag("\n")))(input)?;

    let mut packet = NMDeviceDataPacket {
        mac: device_mac,
        ..Default::default()
    };

    let (input, lines) = context(
        "Получение значений до тега терминатора",
        map_parser(
            take_until1("##"),
            separated_list0(tag("\n"), take_while1(|c| c != '\n')),
        ),
    )(input)?;

    for line in lines {
        parse_packet_body(line, &mut packet)?;
    }

    let (input, _) = tag("##")(input)?;

    packet.name = opt_name.map(|v| v.to_string());

    Ok((input, packet))
}
