use hifitime::Epoch;
use serde::{de, Deserialize, Deserializer, Serialize, Serializer};
use std::fmt;
use std::fmt::Formatter;

#[derive(PartialOrd, PartialEq, Ord, Eq, Clone, Copy, Debug, Default)]
#[repr(transparent)]
pub struct EpochUTC(pub Epoch);

impl From<Epoch> for EpochUTC {
    #[inline]
    fn from(value: Epoch) -> Self {
        EpochUTC(value)
    }
}

impl Serialize for EpochUTC {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        serializer.serialize_f64(self.0.to_unix_seconds())
    }
}

impl<'de> Deserialize<'de> for EpochUTC {
    fn deserialize<D>(deserializer: D) -> Result<EpochUTC, D::Error>
    where
        D: Deserializer<'de>,
    {
        struct EpochVisitor;

        impl de::Visitor<'_> for EpochVisitor {
            type Value = EpochUTC;

            fn expecting(&self, formatter: &mut Formatter) -> fmt::Result {
                formatter.write_str(
                    "a (string, representing a timestamp in decimal or hexadecimal form) or a decimal",
                )
            }

            fn visit_u64<E>(self, v: u64) -> Result<Self::Value, E>
            where
                E: de::Error,
            {
                Ok(Epoch::from_unix_seconds(v as f64).into())
            }

            fn visit_f64<E>(self, v: f64) -> Result<Self::Value, E>
            where
                E: de::Error,
            {
                Ok(Epoch::from_unix_seconds(v).into())
            }

            fn visit_str<E>(self, value: &str) -> Result<Self::Value, E>
            where
                E: de::Error,
            {
                if value.len() == 8 {
                    let bytes = hex::decode(value).map_err(de::Error::custom)?;

                    return Ok(Epoch::from_unix_seconds(u32::from_be_bytes(
                        bytes.as_slice().try_into().unwrap(),
                    ) as f64)
                    .into());
                }
                Ok(
                    Epoch::from_unix_seconds(value.parse().map_err(de::Error::custom)?).into(),
                )
            }
        }

        deserializer.deserialize_any(EpochVisitor)
    }
}
