//! Глобальный модуль для вспомогательных типов и утилит.
//!

mod hifitime_serde;
pub mod uformat;

use phf::phf_map;
use serde::{Deserialize, Deserializer, Serialize, Serializer};
use std::{borrow::Cow, sync::Arc};

pub use hifitime_serde::EpochUTC;

/// Поддерживаемые типы.
///
/// TODO: Решить необходимо ли к данным прикреплять единицы измерения.
/// TODO: Сейчас вообще сомнительно оставлять ли это или нет.
#[allow(dead_code)]
#[derive(Debug, Clone, Hash, PartialEq, Eq)]
#[repr(u64)]
pub enum SupportedUnit {
    Celsius,      // Needs verification > -273.15
    Percentage,   // Needs verification >= 100 && <= 0
    MillimeterHg, // Needs verification
    UVIndex,      // Needs verification
    Boolean,      // Needs verification and possible parsing
    Kbps,
    Volume,
    KWh,
    Amps,
    Volts,
    Watts,
    Seconds,
    Hertz,
}

impl Serialize for SupportedUnit {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        serializer.serialize_str(match self {
            SupportedUnit::Celsius => "C",
            SupportedUnit::Percentage => "%",
            SupportedUnit::MillimeterHg => "mmHg",
            SupportedUnit::UVIndex => "UV",
            SupportedUnit::Boolean => "bool",
            SupportedUnit::Kbps => "Kbps",
            SupportedUnit::Volume => "m3",
            SupportedUnit::Amps => "A",
            SupportedUnit::Volts => "V",
            SupportedUnit::Watts => "W",
            SupportedUnit::Seconds => "s",
            SupportedUnit::KWh => "KWh",
            SupportedUnit::Hertz => "Hz",
        })
    }
}

impl<'de> Deserialize<'de> for SupportedUnit {
    fn deserialize<'a, D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        let color_str = Cow::<'a, str>::deserialize(deserializer)?;
        match STR_TO_UNITS.get(color_str.as_ref()) {
            Some(v) => Ok(v.clone()),
            None => Err(serde::de::Error::custom("Invalid unit")),
        }
    }
}

/// Таблица преобразования текстового представления единиц в значения [SupportedUnit].
static STR_TO_UNITS: phf::Map<&'static str, SupportedUnit> = phf_map! {
    "C" => SupportedUnit::Celsius,
    "%" => SupportedUnit::Percentage,
    "mmHg" => SupportedUnit::MillimeterHg,
    "UV" => SupportedUnit::UVIndex,
    "bool" => SupportedUnit::Boolean,
    "Kbps" => SupportedUnit::Kbps,
    "m3" => SupportedUnit::Volume,
    "A" => SupportedUnit::Amps,
    "V" => SupportedUnit::Volts,
    "W" => SupportedUnit::Watts,
    "s" => SupportedUnit::Seconds,
    "KWh" => SupportedUnit::KWh,
};

pub fn convert_to_arc<T: std::error::Error>(error: T) -> Arc<T> {
    Arc::new(error)
}
