//! Модуль обработки данных с устройств, которые используют старый API.

pub mod qs_parser;

use std::sync::Arc;

use crate::ingest_protocol::NMJsonPacket;
use crate::web_server::app_error::{self, AppError};

use ntex::http::{HttpMessage, StatusCode};
use ntex::util::Bytes;
use ntex::web::types::State;
use ntex::{http, web};
use qs_parser::QSParserError;
use snafu::{ResultExt, Snafu};

use super::NMAppState;

#[derive(Snafu, Debug)]
#[snafu(visibility(pub))]
pub enum Error {
    #[snafu(display("Device not found"))]
    DeviceNotFound {
        mac: String
    },

    #[snafu(display("Time sent with the device is way too behind now"))]
    TimeIsLongBehindNow,

    #[snafu(display("{source}"))]
    QSParser {
        source: QSParserError
    },
}

/// Обработчик данных датчиков с устройств.
///
/// Слушает /post и /get.
/// Для того чтобы пользователям было легче, на оба пути можно отправлять и POST и GET.
///
/// На POST можно отправлять JSON или url-encoded тело, на GET - только через Query String.
pub async fn device_handler(
    request: web::HttpRequest,
    body: Bytes,
    app_state: State<NMAppState>,
) -> Result<web::HttpResponse, AppError> {
    let mut real_body: Option<NMJsonPacket> = None;
    let mut json_error = None;
    let mut query_error = None;

    dbg!(&request.content_type());
    dbg!(&request.query_string());

    if request.method() == http::Method::POST {
        match request.content_type() {
            "application/json" => match serde_json::from_slice::<NMJsonPacket>(body.as_ref()) {
                Ok(json_body) => real_body = Some(json_body),
                Err(error) => json_error = Some(error),
            },
            "application/x-www-form-urlencoded" => {
                let body = std::str::from_utf8(body.as_ref()).context(app_error::Utf8Snafu)?;
                match qs_parser::parse_nm_qs_format(body).await {
                    Ok(qs_body) => {
                        real_body = Some(NMJsonPacket {
                            devices: Vec::from([qs_body]),
                        })
                    }
                    Err(error) => query_error = Some(error),
                }
            }
            _ => {}
        }
    } else if request.method() == http::Method::GET {
        match qs_parser::parse_nm_qs_format(request.query_string()).await {
            Ok(qs_body) => {
                real_body = Some(NMJsonPacket {
                    devices: Vec::from([qs_body]),
                })
            }
            Err(error) => query_error = Some(error),
        }
    }

    if let Some(real_body) = real_body {
        real_body.save_to_db(&app_state.redis_client).await?;
    } else {
        return Err(AppError::UnknownBody {
            json_err: Arc::new(json_error),
            query_error: Arc::new(query_error),
        });
    }

    Ok(web::HttpResponseBuilder::new(StatusCode::OK).finish())
}
