use crate::ingest_protocol::error::Error;
use crate::ingest_protocol::parser::parse_mac_address;
use crate::ingest_protocol::{NMDeviceDataPacket, SensorValue};
use crate::utils::convert_to_arc;
use hifitime::Epoch;
use ntex::util::HashMap;
use rust_decimal::Decimal;
use snafu::{ResultExt, Snafu};
use std::collections::HashSet;
use std::num::ParseFloatError;
use std::str::FromStr;
use std::sync::Arc;

/// В иделае было бы хорошо сделать всё как у [serde_json::Error], но это слишком большая морока
#[allow(dead_code)]
#[derive(Snafu, Clone, Debug)]
#[snafu(visibility(pub))]
pub enum QSParserError {
    #[snafu(display("asd"))]
    SerdeQS {
        #[snafu(source(from(serde_qs::Error, convert_to_arc::<serde_qs::Error>)))]
        source: Arc<serde_qs::Error>
    },
    #[snafu(display("asd"))]
    Parsing {
        context: String
    },

    #[snafu(display("asd"))]
    FloatP {
        #[snafu(source)]
        source: ParseFloatError
    },

    #[snafu(display("failed to parse into decimal"))]
    DecimalParse {
        source: rust_decimal::Error
    },

    #[snafu(display("asd"))]
    NoMAC,
}

impl From<Error<&str>> for QSParserError {
    fn from(value: Error<&str>) -> Self {
        QSParserError::Parsing { context: format!("{:?}", value)}
    }
}

// impl From<serde_qs::Error> for QSParserError {
//     fn from(value: serde_qs::Error) -> Self {
//         QSParserError::SerdeQS(Arc::new(value))
//     }
// }

/// Преобразование оставшихся параметров в urlencoded теле или query string в данные с датчиков
/// [SensorValue].
///
/// Формат: `<SENSOR_MAC>=<SENSOR_VALUE>`.
/// Других данных наподобие названия и времени нет.
pub fn qs_rest_to_values(
    parsed: HashMap<String, String>,
) -> Result<HashSet<SensorValue>, QSParserError> {
    let mut hashset = HashSet::new();

    for (key, value) in parsed {
        hashset.insert(SensorValue {
            mac: key,
            value: Decimal::from_str(value.as_str()).context(DecimalParseSnafu{})?,

            time: None,
            unit: None,
            name: None,
        });
    }

    Ok(hashset)
}

pub fn parse_decimal_if_exists(
    parsed: &mut HashMap<String, String>,
    key: &str,
) -> Result<Option<Decimal>, QSParserError> {
    if let Some(unwrapped_value) = parsed.remove(key) {
        Ok(Some(Decimal::from_str(unwrapped_value.as_str()).context(DecimalParseSnafu{})?))
    } else {
        Ok(None)
    }
}

pub fn parse_epoch_if_exists(
    parsed: &mut HashMap<String, String>,
    key: &str,
) -> Result<Option<Epoch>, QSParserError> {
    if let Some(unwrapped_value) = parsed.remove(key) {
        Ok(Some(Epoch::from_unix_seconds(unwrapped_value.parse().context(FloatPSnafu{})?)))
    } else {
        Ok(None)
    }
}

pub async fn parse_nm_qs_format(input: &str) -> Result<NMDeviceDataPacket, QSParserError> {
    let mut parsed: HashMap<String, String> = serde_qs::from_str(input).context(SerdeQSSnafu{})?;

    let (_, device_mac) = if let Some(id) = parsed.get("ID") {
        parse_mac_address(id)?
    } else {
        return Err(QSParserError::NoMAC);
    };

    parsed.remove("ID");

    let device_data = NMDeviceDataPacket {
        mac: device_mac,
        name: parsed.remove("name").map(|v| v.to_owned()),
        owner: parsed.remove("owner").map(|v| v.to_owned()),
        lat: parse_decimal_if_exists(&mut parsed, "lat")?,
        lon: parse_decimal_if_exists(&mut parsed, "lon")?,
        alt: parse_decimal_if_exists(&mut parsed, "alt")?,
        time: parse_epoch_if_exists(&mut parsed, "time")?.map(|v| v.into()),
        // TODO: Выяснить можно ли передавать команды по QS и можно ли их отличить от
        //       маков и значений сенсоров.
        commands: None,

        values: qs_rest_to_values(parsed)?,
    };

    Ok(device_data)
}
